from __future__ import annotations

from functools import partial
from itertools import count
from math import e, log, pi, sin, sqrt
from typing import Any, Iterable, Literal

from vsexprtools import ExprOp, aka_expr_available, norm_expr, norm_expr_planes, ExprVars
from vstools import (
    ColorRange, ConvMode, CustomIndexError, CustomValueError, PlanesT, StrList, VSFunction, check_ref_clip,
    check_variable, core, disallow_variable_format, disallow_variable_resolution, flatten, get_peak_value, get_y, join,
    normalize_planes, scale_value, split, vs, get_video_format
)

from .blur import box_blur, gauss_blur

__all__ = [
    'replace_low_frequencies',
    'diff_merge', 'lehmer_diff_merge'
]


@disallow_variable_format
@disallow_variable_resolution
def replace_low_frequencies(
    flt: vs.VideoNode, ref: vs.VideoNode, LFR: float, DCTFlicker: bool = False,
    planes: PlanesT = None, mode: ConvMode = ConvMode.SQUARE
) -> vs.VideoNode:
    assert check_variable(flt, replace_low_frequencies)
    check_ref_clip(flt, ref, replace_low_frequencies)

    planes = normalize_planes(flt, planes)
    work_clip, *chroma = split(flt) if planes == [0] else (flt, )
    assert check_variable(work_clip, replace_low_frequencies)

    ref_work_clip = get_y(ref) if work_clip.format.num_planes == 1 else ref

    LFR = max(LFR or (300 * work_clip.width / 1920), 50)

    freq_sample = max(work_clip.width, work_clip.height) * 2    # Frequency sample rate is resolution * 2 (for Nyquist)
    k = sqrt(log(2) / 2) * LFR                                  # Constant for -3dB
    LFR = freq_sample / (k * 2 * pi)                            # Frequency Cutoff for Gaussian Sigma
    sec0 = sin(e) + 0.1

    sec = scale_value(sec0, 8, work_clip.format.bits_per_sample, range_out=ColorRange.FULL)

    expr = 'x y - z + '

    if DCTFlicker:
        expr += f'y z - D! y z = swap dup D@ 0 = 0 D@ 0 < -1 1 ? ? {sec} * + ?'

    flt_blur = gauss_blur(work_clip, LFR, None, mode)
    ref_blur = gauss_blur(ref_work_clip, LFR, None, mode)

    final = norm_expr(
        [work_clip, flt_blur, ref_blur], expr, planes, force_akarin='vsrgtools.replace_low_frequencies'
    )

    if not chroma:
        return final

    return join([final, *chroma], flt.format.color_family)


def diff_merge(
    *_clips: vs.VideoNode | Iterable[vs.VideoNode],
    filter: VSFunction | list[VSFunction] = partial(box_blur, radius=1, passes=3),
    operator: Literal['<', '>'] = '>', abs_diff: bool = True, **kwargs: Any
) -> vs.VideoNode:

    clips = list[vs.VideoNode](flatten(_clips))  # type: ignore
    n_clips = len(clips)

    if not aka_expr_available and n_clips > 13:
        raise CustomIndexError(f'Too many clips passed! ({n_clips})', diff_merge)
    elif n_clips < 2:
        raise CustomIndexError(f'You must pass at lest two clips! ({n_clips})', diff_merge)

    if not filter:
        raise CustomValueError('You must pass at least one filter!', diff_merge)

    if not isinstance(filter, list):
        filter = [filter]

    if (lf := len(filter)) < n_clips:
        filter.extend(filter[-1:] * (n_clips - lf))
    elif lf > n_clips:
        filter = filter[:n_clips]

    blurs = [filt(clip, **kwargs) for filt, clip in zip(filter, clips)]

    expr_string = ''
    n_clips = len(clips)

    for src, flt in zip(ExprVars(n_clips), ExprVars(n_clips, n_clips * 2)):
        expr_string += f'{src} {flt} - {abs_diff and "abs" or ""} {src.upper()}D! '

    for i, src, srcn in zip(count(), ExprVars(n_clips), ExprVars(1, n_clips)):
        expr_string += f'{src.upper()}D@ {srcn.upper()}D@ {operator} {src} '

        if i == n_clips - 2:
            expr_string += f'{srcn} '

    expr_string += '? ' * (n_clips - 1)

    return norm_expr([*clips, *blurs], expr_string, force_akarin='vsrgtools.diff_merge')


def lehmer_diff_merge(
    *_clips: vs.VideoNode | Iterable[vs.VideoNode],
    filter: VSFunction | list[VSFunction] = partial(box_blur, radius=3, passes=2),
    planes: PlanesT = None
) -> vs.VideoNode:
    """
    Merge multiple sources.

    :param clips:           Clips to be merged together.
                            If using multiple sources, first should be the "best" one,
                            then the other a lowpassed or worse source overall.
                            If the clips are filtered, first is src and
                            others the denoised/debanded/whatever.
    :param filter:          Filter to be applied to clips to isolate high frequencies.
                            If a list, each filter will be mapped to corresponding clip at same index.
                            If not enough filters are passed, the last one will be reused.
    :param planes:          Planes to be processed.

    :return:                Merged clips.
    """

    clips = list[vs.VideoNode](flatten(_clips))  # type: ignore
    n_clips = len(clips)

    if not aka_expr_available and n_clips > 2:
        raise CustomIndexError('You can pass at most 2 clips without akarin plugin!', lehmer_diff_merge, n_clips)
    elif n_clips < 2:
        raise CustomIndexError('You must pass at lest two clips!', lehmer_diff_merge, n_clips)

    if not filter:
        raise CustomValueError('You must pass at least one filter!', lehmer_diff_merge)

    formats = {get_video_format(clip).id for clip in clips}

    if len(formats) > 1:
        raise CustomValueError('All clips must have the same format!', lehmer_diff_merge)

    if not isinstance(filter, list):
        filter = [filter]

    if (lf := len(filter)) < n_clips:
        filter.extend(filter[-1:] * (n_clips - lf))
    elif lf > n_clips:
        filter = filter[:n_clips]

    blurred_clips = [filt(clip) for filt, clip in zip(filter, clips)]

    if aka_expr_available:
        counts = range(n_clips)

        clip_vars, blur_vars = ExprVars(n_clips), ExprVars(n_clips, n_clips * 2)

        n_op = n_clips - 1

        expr = StrList([
            [f'{clip} {blur} - D{i}!' for i, clip, blur in zip(counts, clip_vars, blur_vars)],
        ])

        for y in range(2):
            expr.extend([
                [f'D{i}@ {3 - y} pow' for i in counts],
                ExprOp.ADD * n_op, f'P{y + 1}!'
            ])

        expr.extend([
            'P2@ 0 = 0 P1@ P2@ / ?',
            blur_vars, ExprOp.ADD * n_op,
            n_clips, ExprOp.DIV, ExprOp.ADD
        ])

        return norm_expr([*clips, *blurred_clips], expr, planes)

    src, flt = clips
    src_high, flt_high = blurred_clips

    peak = get_peak_value(src)

    src_high_diff = src.std.MakeDiff(src_high, planes)
    flt_high_diff = flt.std.MakeDiff(flt_high, planes)

    flt_high_diff = core.std.Expr(
        [src_high_diff, flt_high_diff],
        norm_expr_planes(
            src, f'x {peak} - 3 pow y {peak} - 3 pow + x {peak} - 2 pow y {peak} - 2 pow 0.0001 + + / {peak} +', planes
        )
    )

    return src.std.MakeDiff(src_high_diff, planes).std.MergeDiff(flt_high_diff, planes)



from __future__ import annotations

from math import log2
from vsexprtools import norm_expr
from vstools import vs, check_variable, get_sample_type, VSFunction, check_ref_clip, PlanesT, normalize_planes, core, CustomTypeError

from .blur import gauss_blur, min_blur
from .limit import limit_filter
from .util import normalize_radius, wmean_matrix, mean_matrix

__all__ = [
    'unsharpen',
    'unsharp_masked',
    'limit_usm'
]


def unsharpen(
    clip: vs.VideoNode, strength: float = 1.0, sigma: float | list[float] = 1.5,
    prefilter: vs.VideoNode | VSFunction | None = None
) -> vs.VideoNode:
    assert check_variable(clip, unsharpen)

    if callable(prefilter):
        ref = prefilter(clip)
    else:
        ref = prefilter

    check_ref_clip(clip, ref)

    den = ref or clip
    blur = gauss_blur(den, sigma)

    unsharp = norm_expr([den, blur], f'x y - {strength} * x +', 0)

    if ref is not None:
        return unsharp

    return unsharp.std.MergeDiff(clip.std.MakeDiff(den))


def unsharp_masked(
    clip: vs.VideoNode, radius: int | list[int] = 1, strength: float = 100.0, planes: PlanesT = None
) -> vs.VideoNode:
    planes = normalize_planes(clip, planes)

    if isinstance(radius, list):
        return normalize_radius(clip, unsharp_masked, radius, planes, strength=strength)

    strength = max(1e-6, min(log2(3) * strength / 100, log2(3)))

    weight = 0.5 ** strength / ((1 - 0.5 ** strength) * 0.5)

    if get_sample_type(clip) == 0:
        all_matrices = [[x * 1.0] for x in range(1, 1024)]

        for x in range((2 ** 10) - 1):
            while len(all_matrices[x]) < radius * 2 + 1:
                all_matrices[x].append(all_matrices[x][-1] / weight)

        error = list(map(sum, ([abs(x - round(x)) for x in matrix[1:]] for matrix in all_matrices)))
        matrix = list(map(float, map(round, all_matrices[error.index(min(error))])))
    else:
        matrix = [1.0]

        while len(matrix) < radius * 2 + 1:
            matrix.append(matrix[-1] / weight)

    if radius == 1:
        indices = (2, 1, 2, 1, 0, 1, 2, 1, 2)
    else:
        indices = (4, 3, 2, 3, 4, 3, 2, 1, 2, 3, 2, 1, 0, 1, 2, 3, 2, 1, 2, 3, 4, 3, 2, 3, 4)

    matrix = [matrix[x] for x in indices]

    blurred = clip.std.Convolution(matrix, planes=planes)

    return clip.std.MergeDiff(clip.std.MakeDiff(blurred, planes=planes), planes=planes)


def limit_usm(
    clip: vs.VideoNode, blur: int | vs.VideoNode | VSFunction = 1, thr: int | tuple[int, int] = 3,
    elast: float = 4.0, bright_thr: int | None = None, planes: PlanesT = None
) -> vs.VideoNode:
    """Limited unsharp_masked."""

    if callable(blur):
        blurred = blur(clip)
    elif isinstance(blur, vs.VideoNode):
        blurred = blur
    elif blur <= 0:
        blurred = min_blur(clip, -blur, planes)
    elif blur == 1:
        blurred = clip.std.Convolution(wmean_matrix, planes=planes)
    elif blur == 2:
        blurred = clip.std.Convolution(mean_matrix, planes=planes)
    else:
        raise CustomTypeError("'blur' must be an int, clip or a blurring function!", limit_usm, blur)

    sharp = norm_expr([clip, blurred], 'x dup y - +', planes)

    return limit_filter(sharp, clip, thr=thr, elast=elast, bright_thr=bright_thr)



from __future__ import annotations

from enum import auto
from typing import Sequence

from vstools import CustomIntEnum

__all__ = [
    'LimitFilterMode',
    'RemoveGrainMode', 'RemoveGrainModeT',
    'RepairMode', 'RepairModeT',
    'VerticalCleanerMode', 'VerticalCleanerModeT'
]


class LimitFilterModeMeta:
    force_expr = True


class LimitFilterMode(LimitFilterModeMeta, CustomIntEnum):
    """Two sources, one filtered"""
    SIMPLE_MIN = auto()
    SIMPLE_MAX = auto()
    """One source, two filtered"""
    SIMPLE2_MIN = auto()
    SIMPLE2_MAX = auto()
    DIFF_MIN = auto()
    DIFF_MAX = auto()
    """One/Two sources, one filtered"""
    CLAMPING = auto()

    @property
    def op(self) -> str:
        return '<' if 'MIN' in self._name_ else '>'

    def __call__(self, force_expr: bool = True) -> LimitFilterMode:
        self.force_expr = force_expr

        return self


class RemoveGrainMode(CustomIntEnum):
    NONE = 0
    MINMAX_AROUND1 = 1
    MINMAX_AROUND2 = 2
    MINMAX_AROUND3 = 3
    MINMAX_MEDIAN = 4
    EDGE_CLIP_STRONG = 5
    EDGE_CLIP_MODERATE = 6
    EDGE_CLIP_MEDIUM = 7
    EDGE_CLIP_LIGHT = 8
    LINE_CLIP_CLOSE = 9
    MIN_SHARP = 10
    SQUARE_BLUR = 11
    BOB_TOP_CLOSE = 13
    BOB_BOTTOM_CLOSE = 14
    BOB_TOP_INTER = 15
    BOB_BOTTOM_INTER = 16
    MINMAX_MEDIAN_OPP = 17
    LINE_CLIP_OPP = 18
    CIRCLE_BLUR = 19
    BOX_BLUR = 20
    OPP_CLIP_AVG = 21
    OPP_CLIP_AVG_FAST = 22
    EDGE_DEHALO = 23
    EDGE_DEHALO2 = 24
    MIN_SHARP2 = 25
    SMART_RGC = 26
    SMART_RGCL = 27
    SMART_RGCL2 = 28


RemoveGrainModeT = int | RemoveGrainMode | Sequence[int | RemoveGrainMode]


class RepairMode(CustomIntEnum):
    NONE = 0
    MINMAX_SQUARE1 = 1
    MINMAX_SQUARE2 = 2
    MINMAX_SQUARE3 = 3
    MINMAX_SQUARE4 = 4
    LINE_CLIP_MIN = 5
    LINE_CLIP_LIGHT = 6
    LINE_CLIP_MEDIUM = 7
    LINE_CLIP_STRONG = 8
    LINE_CLIP_CLOSE = 9
    MINMAX_SQUARE_REF_CLOSE = 10
    MINMAX_SQUARE_REF1 = 11
    MINMAX_SQUARE_REF2 = 12
    MINMAX_SQUARE_REF3 = 13
    MINMAX_SQUARE_REF4 = 14
    CLIP_REF_RG5 = 15
    CLIP_REF_RG6 = 16
    CLIP_REF_RG17 = 17
    CLIP_REF_RG18 = 18
    CLIP_REF_RG19 = 19
    CLIP_REF_RG20 = 20
    CLIP_REF_RG21 = 21
    CLIP_REF_RG22 = 22
    CLIP_REF_RG23 = 23
    CLIP_REF_RG24 = 24
    CLIP_REF_RG26 = 26
    CLIP_REF_RG27 = 27
    CLIP_REF_RG28 = 28


RepairModeT = int | RepairMode | Sequence[int | RepairMode]


class VerticalCleanerMode(CustomIntEnum):
    NONE = 0
    MEDIAN = 1
    PRESERVING = 2


VerticalCleanerModeT = int | VerticalCleanerMode | Sequence[int | VerticalCleanerMode]
